/* Copyright (c) 2006-2015 Polybus Systems Corp., Westford, MA, USA */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "sys_basher.h"
#include <pthread.h>
#include <sys/mman.h>

int main(int argc, char *argv[])
{
  pthread_t *thread_st=0;
  exp_data expected;
  result_summary *results=0;
  result_summary *final_results=0;
  run_summary ran;
  UInt64 *seed=0;
  SInt32 MAX_THREADS=0;
  long errors,loops=1,i,j,k,value,threads=4,vnum=0,first_thread=1;
  UInt64 align;
  UInt64 select=0;
  SInt64 test;
  SInt64 *base=0,*cindx=0;
  SInt64 limit,sizepls,words,long_limit,indx,offset,addr_limit,total_minus=0,arraysize=0;
  SInt64 totalram,freeram,totalswap,freeswap,sharedram,bufferram,pagesize;
  SInt64 dutycycle=100;
  double minutes=0,stime=0,etime=0,ctime=0,dutyscale=0,ttime=0,tstart=0,tend=0;
  double size;
  my_sysinfo sys_info;
  char isPath;
  file_info finfo;
  thread_data *t_data=0;
  void *mem;
  double temps[64];
  double max_temps[64],min_temps[64];
  char temp_names[64][64];
  char report_name[512],string[1024];
  long num_temps=0,thread_base=0,thread_limit=0,numCPUs=0;
  Boolean stop_on_error = 0;
  Boolean use_total = 0;

  numCPUs = sysconf(_SC_NPROCESSORS_ONLN);
  MAX_THREADS = ((numCPUs +1) * (numCPUs/2)) + 1;

  sizepls = sizeof(pthread_t) * (MAX_THREADS + 1L);
  thread_st = (pthread_t *)malloc(sizepls);

  sizepls = sizeof(result_summary) * (MAX_THREADS + 1L);
  results = (result_summary *)malloc(sizepls);
  final_results = (result_summary *)malloc(sizepls);

  //  printf("MAX_THREADS: %ld,sizepls: %lld\n",MAX_THREADS,sizepls);

  sizepls = sizeof(thread_data) * (MAX_THREADS + 1L);
  t_data = (thread_data *)malloc(sizepls);

  sizepls = sizeof(UInt64) * (MAX_THREADS + 1L);
  seed = (UInt64 *)malloc(sizepls);

  strcpy(report_name,"sys_basher");

  for(i=0;i<MAX_THREADS;i++)
    {
      seed[i] = ((UInt64)i + 1LL) * 0x0123456789ABCDEFLL;
    }

  threads = numCPUs;

  sync();
  sync();
  mySysInfo(&sys_info);

  finfo.num_paths = 0;
  i = 1;
  select = 0;
  dutycycle = 100;

  while(argv[i])
    {
      if(!getSwitch(argv[i],string))
	{
	  i++;
	}
      else
	{

	  if((string[0] == 's') && (string[1] == 'w'))
	    {
	      //swap
	      select |= SWAP_ONLY;
	      i++;
	    }
	  else if((string[0] == 'f') && ((string[1] == 'a') || (string[1] == 0)))
	    {
	      //fast
	      select |= BAND_ONLY | DISK_ONLY;
	      i++;
	    }
	  else if((string[0] == 'd') && ((string[1] == 'i') || (string[1] == 0)))
	    {
	      select |= DISK_ONLY;
	      i++;
	    }
	  else if((string[0] == 'd') && (string[1] == 'm') && (string[2] == 'i') && (string[3] == 0))
	    {
	      i++;
	      openDMIFile(argv[i]);
	      i++;
	    }
	  else if((string[0] == 'f') && (string[1] == 'p'))
	    {
	      select |= FP_ONLY;
	      i++;
	    }
	  else if(string[0] == 'i')
	    {
	      //int
	      select |= INT_ONLY;
	      i++;
	    }
	  else if((string[0] == 'm') && (string[1] == 'b'))
	    {
	      select |= BAND_ONLY;
	      i++;
	    }
	  else if((string[0] == 'm') && ((string[1] == 'e') || (string[1] == 0)))
	    {
	      i++;
	      if(getSwitchArg(argv[i],string))
		{
		  i++;
		  j=0;
		  while(string[j])
		    {
		      switch(string[j])
			{
			case 'a': select |= ADDR_EQ_ONLY; break;
			case 'b': select |= BANK_ONLY; break;
			case 'p': select |= PATTERN_ONLY; break;
			case 'r': select |= RANDOM_ONLY; break;
			case 'w': select |= WALKING_ONLY; break;
			case 'v': select |= REV_ADDR_ONLY; break;
			default: break;
			}
		      j++;
		    }
		}
	      else
		{
		  select |= ADDR_EQ_ONLY | BANK_ONLY | PATTERN_ONLY | RANDOM_ONLY | WALKING_ONLY | REV_ADDR_ONLY;
		}
	      
	    }
	  else if((string[0] == 'v') && ((string[1] == 0) || (string[3] == 'b')))
	    {
	      setVerbose(1);
	      i++;
	    }
	  else if((string[0] == 'v') && ((string[1] == 'v') && ((string[2] == 0) || (string[4] == 'b'))))
	    {
	      setVerbose(2);
	      i++;
	    }
	  else if((string[0] == 'v') && (string[1] == 'v') && ((string[2] == 'v') && ((string[3] == 0) || (string[5] == 'b'))))
	    {
	      setVerbose(3);
	      i++;
	    }
	  else if(string[0] == 'l')
	    {
	      //loops
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  if(value != 0L)
		    {
		      loops = value;
		      i++;
		    }
		}
	    }
	  else if((string[0] == 'h') && (string[1] == 'o'))
	    {
	      //hours
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  if(value != 0L)
		    {
		      minutes = value * 60.0;
		      i++;
		    }
		}
	    }
	  else if((string[0] == 'h') && ((string[1] == 'e') || (string[1] == 0)))
	    {
	      explainIt();
	      exit(0);
	    }
	  else if((string[0] == 's') && (string[1] == 't'))
	    {
	      stop_on_error = 1;
	      i++;
	    }
	  else if((string[0] == 'v') && (string[1] == 'e'))
	    {
	      getVersion();
	      exit(0);
	    }
	  else if((string[0] == 'm') && (string[1] == 'i'))
	    {
	      //minutes
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  if(value != 0L)
		    {
		      minutes = value;
		      i++;
		    }
		}
	    }
	  else if((string[0] == 'd') && (string[1] == 'u'))
	    {
	      //dutycycle
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  if(value != 0L)
		    {
		      dutycycle = value;
		      i++;
		    }
		}
	    }
	  else if((string[0] == 'p') && ((string[1] == 'a') || (string[1] == 0)))
	    {
	      //path
	      i++;
	      isPath = 0;
	      if(argv[i]) isPath = argv[i][0] != '-';
	      while(isPath)
		{
		  if(finfo.num_paths < 31)
		    {
		      pathCopy(finfo.path[finfo.num_paths],argv[i],511);
		      finfo.num_paths++;
		    }
		  i++;
		  isPath = 0;
		  if(argv[i]) isPath = argv[i][0] != '-';
		}
	    }
	  else if(string[0] == 'r')
	    {
	      //report
	      i++;
	      if(argv[i] && (argv[i][0] != '-'))
		{
		  pathCopy(report_name,argv[i],510);
		  i++;
		}
	    }
	  else if((string[0] == 't') && ((string[1] == 'o') || (string[1] == 0)))
	    {
	      //totalmem
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  total_minus = value * 0x100000LL;
		  use_total = 1;
		  i++;
		}
	    }
	  else if((string[0] == 't') && (string[1] == 's'))
	    {
	      // Thread Start
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  if(value < 1)
		    {
		      first_thread = 1;
		    }
		  else if(value < MAX_THREADS)
		    {
		      first_thread = value;
		    }
		  else
		    {
		      first_thread = MAX_THREADS;
		    }
		  i++;
		}
	      
	    }
	  else if((string[0] == 't') && (string[1] == 'h'))
	    {
	      //threads
	      i++;
	      if(argv[i])
		{
		  value = strToInt(argv[i]);
		  if(value != 0L)
		    {
		      if(value < MAX_THREADS)
			{
			  threads = value;
			}
		      else
			{
			  threads = MAX_THREADS;
			}
		      i++;
		    }
		}
	    }
	  else
	    {
	      i++;
	    }

	}
    }

  openTranscript(report_name);
  openDMIFile("/var/sys_basher/bios.bin");
  //  openDMIFile("bios.bin");
  totalram = sys_info.totalram * sys_info.mem_unit;
  freeram = sys_info.freeram * sys_info.mem_unit;
  totalswap = sys_info.totalswap * sys_info.mem_unit;
  freeswap = sys_info.freeswap * sys_info.mem_unit;
  sharedram = sys_info.sharedram * sys_info.mem_unit;
  bufferram = sys_info.bufferram * sys_info.mem_unit;
  pagesize = getpagesize();

  sprintf(string,"sys_basher version %s\nNumber of CPUs: %ld\n",getRevision(),numCPUs);
  printTranscript(string,TRUE);
  mprint(12L,"Total RAM",totalram);
  mprint(12L,"Free RAM",freeram);
  mprint(12L,"Total Swap",totalswap);
  mprint(12L,"Free Swap",freeswap);
  mprint(12L,"Buffer RAM",bufferram);
  mprint(12L,"Shared RAM",sharedram);
  mprint(12L,"Page Size",pagesize);

  printCPUFrequency("\nMin CPU Frequencies\n","cpuinfo_min_freq",FALSE);
  printCPUFrequency("\nMax CPU Frequencies\n","cpuinfo_max_freq",FALSE);

  printDMIEntries();

  for(i=0;i<64;i++)
    {
      max_temps[i] = 0;
      min_temps[i] = 0;
      temps[i] = 0;
      temp_names[i][0] = 0;
    }

  printTime();
  if(getTemps(&num_temps,temp_names,temps))
    {
      printSensors(0L,num_temps,temp_names,temps,"Initial");
      minMaxTemps(num_temps,temps,min_temps,max_temps,TRUE);
      printCPUFrequency("\nCurrent CPU Frequencies\n","cpuinfo_cur_freq",TRUE);
    }

  sync();

  clrExpected(&expected);

  for(i=0;i<MAX_THREADS;i++)
    {
      clrResults(&results[i]);
    }

  for(i=0;i<MAX_THREADS;i++)
    {
      clrResults(&final_results[i]);
    }

  if(finfo.num_paths == 0)
    {
      pathCopy(finfo.path[finfo.num_paths],".",511);
      finfo.num_paths++;
    }

 
  if(use_total)
    {
      limit = totalram;
      limit -= total_minus;
      if(limit < 0) limit = totalram/8L;
    }
  else
    {
      if(totalram > 0x40000000LL)
	{
	  limit = totalram - 0x10000000LL;
	}
      else
	{
	  limit = totalram * 4LL;
	  limit /= 5LL;
	}
      if(freeram < limit) limit = freeram;
    }
  limit &= 0xFFFFFFFFFFFF0000LL;

  sizepls = totalram;
  sizepls *= 5LL;
  sizepls /= 4LL;

  sizepls &= 0xFFFFFFFFFFFF0000LL;

  createDirs(MAX_THREADS,&finfo);

  if(dutycycle >= 100L)
    {
      dutyscale = 0.0;
    }
  else
    {
      dutyscale = dutycycle;
      dutyscale = 100.0/dutyscale;
      dutyscale = dutyscale - 1.0;
    }

  if((select & SWAP_ONLY) == 0) sizepls = limit + 0x10000L;

  mem = malloc(sizepls);
  while((mem == NULL) && (sizepls > 0x1000000))
    {
      sizepls *= 7LL;
      sizepls /= 8LL;
      sizepls &= 0xFFFFFFFFFFFF0000LL;
      mem = malloc(sizepls);
    }
  if(mem == NULL)
    {
      printTranscript("Can't allocate enough memory, please close all open applications\n",TRUE);
      
      deleteDirs(MAX_THREADS,&finfo);
      exit(1);
    }
  arraysize = sizepls;

  align = (UInt64)mem;

  sizepls -= 0x10000LL;

  if((align & 0x0FFFFLL) != 0)
    {
      align += 0x0FFFFLL;
      align &= 0xFFFFFFFFFFFF0000LL;
      sizepls -= 0x10000LL;
    }

  if(limit > sizepls) limit = sizepls;
  
  size = limit;
  commaNumber(size,string);
  mprint(18L,"Allocated Memory",size);
  base = (void *)align;
  limit &= 0xFFFFFFFFFFFF0000LL;
  if((totalram - 0x8000000LL) > sizepls)
    {
      // If there is 128M available after the test memory has been allocated then lock the memory to prevent paging
      // If there is less than paging is permitted
      mlock(base,limit);
      cindx = base;
      for(i=0;i<limit;i+=8LL)
	{
	  *cindx++ = 0LL;
	}
    }

  openPageMap();
  //  translatePages((UInt64)base,64);

  errors = 0;
  minutes *= 60.0;

  stime = getTime();
  stime /= 1000000.0;
  etime = stime + minutes;
  i=0;

  flushTranscript();
  //-- Initialize the FP tests
  words = limit/8LL;
  offset = words;
  offset &= 0xFFFFFFFFFFFFE000LL;
  addr_limit = offset * 8LL;
  long_limit = sizepls;
  long_limit &= 0xFFFFFFFFFFFFE000LL;
  indx = 0;
  t_data[0].base = &base[indx];
  t_data[0].limit = addr_limit;
  t_data[0].long_limit = long_limit;
  t_data[0].errors = 0;
  t_data[0].num = 1;
  t_data[0].vnum = 1;
  t_data[0].num_threads = 1;
  t_data[0].finfo = &finfo;
  t_data[0].firsttime = 1;
  t_data[0].expected = &expected;
  t_data[0].rslt_summary = &results[0];
  t_data[0].seed = &seed[0];
  t_data[0].verbose = getVerbose();
  if(testEnb(FP_TST,select))
    {
      t_data[0].test = FP_TST;
      runTests(&t_data[0]);
      flushTranscript();
    }
  if(testEnb(INT_TST,select))
    {
      t_data[0].test = INT_TST;
      runTests(&t_data[0]);
      flushTranscript();
    }
  if(first_thread > threads) first_thread = threads;
  i=0;
  while((i < loops) || (ctime < etime))
    {
      sprintf(string,"Loop:%ld of %ld\n",i,loops);
      printTranscript(string,FALSE);
      words = limit/8LL;

      for(test=0;test<NUM_TESTS;test++)
	{
	  if(testEnb(test,select))
	    {
	      thread_base = 0;
	      for(k=first_thread;k<=threads;k++)
		{
		  thread_limit = thread_base + k;
		  offset = words/(SInt64)k;
		  offset &= 0xFFFFFFFFFFFFE000LL;
		  addr_limit = offset * 8LL;
		  long_limit = sizepls/(SInt64)k;
		  long_limit &= 0xFFFFFFFFFFFFE000LL;

		  tstart = getTime();
		  if(thread_limit < MAX_THREADS)
		    {
		      for(j=thread_base;j<thread_limit;j++)
			{
			  
			  indx = j - thread_base;
			  vnum = indx+1;
			  sprintf(string,"Create thread: %ld of %ld\n",(long)vnum,k);
			  printTranscript(string,TRUE);

			  indx *= offset;
			  t_data[j].base = &base[indx];
			  t_data[j].limit = addr_limit;
			  t_data[j].long_limit = long_limit;
			  t_data[j].errors = 0;
			  t_data[j].vnum = vnum;
			  t_data[j].num = j + 1;
			  t_data[j].num_threads = k;
			  t_data[j].test = test;
			  t_data[j].finfo = &finfo;
			  t_data[j].firsttime = 0;
			  t_data[j].expected = &expected;
			  t_data[j].rslt_summary = &results[j];
			  t_data[j].seed = &seed[j];
			  t_data[j].verbose = getVerbose();
			  pthread_create(&thread_st[j], NULL, runTests, (void *)&t_data[j]);
			}
		      for(j=thread_base;j<thread_limit;j++)
			{
			  pthread_join(thread_st[j],NULL);
			}
		      for(j=thread_base;j<thread_limit;j++)
			{
			  errors += t_data[j].errors;
			}
		      flushTranscript();
		      if(stop_on_error)
			{
			  if(errors != 0)
			    {
			      genReports(ctime,
					 stime,
					 report_name,
					 results,
					 final_results,
					 &ran,
					 &finfo,
					 threads,
					 thread_base,
					 thread_limit,
					 errors,
					 totalram,
					 freeram,
					 totalswap,
					 freeswap,
					 sharedram,
					 bufferram,
					 pagesize,
					 num_temps,
					 max_temps,
					 min_temps,
					 temp_names,
					 mem,
					 arraysize,
					 MAX_THREADS
					 );
			      exit(1);
			    }
			}
		    }
		  thread_base = thread_limit + 1;
		  tend = getTime();
		  ttime = tend - tstart;
		  cpuRest(ttime,dutyscale);
		}

	      printTime();
	      if(getTemps(&num_temps,temp_names,temps))
		{
		  printSensors(0L,num_temps,temp_names,temps,0L);
		  minMaxTemps(num_temps,temps,min_temps,max_temps,FALSE);
		  printCPUFrequency("\nCurrent CPU Frequencies\n","cpuinfo_cur_freq",TRUE);
		}
	    }
	}

      i++;
      ctime = getTime();
      ctime /= 1000000.0;
      if(ctime < etime)
	{
	  sprintf(string,"Minutes Remaining: %0.0f\n",etime - ctime);
	  printTranscript(string,TRUE);
	}
      if(i < loops) 
	{
	  sprintf(string,"Iterations Remaining: %ld\n",loops - i);
	  printTranscript(string,TRUE);
	}
    }
  genReports(ctime,
	     stime,
	     report_name,
	     results,
	     final_results,
	     &ran,
	     &finfo,
	     threads,
	     thread_base,
	     thread_limit,
	     errors,
	     totalram,
	     freeram,
	     totalswap,
	     freeswap,
	     sharedram,
	     bufferram,
	     pagesize,
	     num_temps,
	     max_temps,
	     min_temps,
	     temp_names,
	     mem,
	     arraysize,
	     MAX_THREADS
	     );

  free(thread_st);
  free(results);
  free(final_results);
  free(t_data);
  free(seed);

  if(errors == 0)
    {
      exit(0);
    }
  else
    {
      exit(1);
    }

}

void genReports(double ctime,
		double stime,
		char *report_name,
		result_summary *results,
		result_summary *final_results,
		run_summary *ran,
		file_info *finfo,
		long threads,
		long thread_base,
		long thread_limit,
		long errors,
		SInt64 totalram,
		SInt64 freeram,
		SInt64 totalswap,
		SInt64 freeswap,
		SInt64 sharedram,
		SInt64 bufferram,
		SInt64 pagesize,
		long num_temps,
		double max_temps[64],
		double min_temps[64],
		char temp_names[64][64],
		void *mem,
		SInt64 arraysize,
		SInt32 MAX_THREADS
		)
{
  long i,k;
  double etime;
  char string[1024];
  FILE *fp=0,*csv=0,**csvt=0;
  
  etime = ctime - stime;

  i = sizeof(FILE *) * (MAX_THREADS + 1L);
  csvt = (FILE **)malloc(i);
  
  for(i=0;i<=MAX_THREADS;i++)
    {
      csvt[i] = 0;
    }

  if(report_name[0] != 0)
    {
      sprintf(string,"%s.rpt",report_name);
      fp = fopen(string,"w");
      sprintf(string,"%s_summary.csv",report_name);
      csv = fopen(string,"w");
      sprintf(string,"%s_all_threads.csv",report_name);
      csvt[0] = fopen(string,"w");
      for(i=1;i<=threads;i++)
	{
	  sprintf(string,"%s_%ld_threads.csv",report_name,i);
	  csvt[i] = fopen(string,"w");
	}
    }


  averageAllResults(0,MAX_THREADS,results,&final_results[0],ran,MAX_THREADS);

  thread_base = 0;
  for(k=1;k<=threads;k++)
    {
      thread_limit = thread_base + k;
      if(thread_limit < MAX_THREADS)
	{
	  averageAllResults(thread_base,thread_limit,results,&final_results[k],ran,MAX_THREADS);
	}
      thread_base = thread_limit + 1;
    }
  if(errors == 0L)
    {
      if(fp)fprintf(fp,"PASSED, ran for %0.2f seconds\n\n",etime);
      if(csv)fprintf(csv,"PASSED,ran for %0.2f seconds\n\n",etime);
    }
  else
    {
      if(fp)fprintf(fp,"FAILED, %ld errors\n\n",errors);
      if(csv)fprintf(csv,"FAILED,%ld errors\n\n",errors);
    }

  printRamSize(fp,csv,"Total RAM",totalram);
  printRamSize(fp,csv,"Free RAM",freeram);
  printRamSize(fp,csv,"Total Swap",totalswap);
  printRamSize(fp,csv,"Free Swap",freeswap);
  printRamSize(fp,csv,"Buffer RAM",bufferram);
  printRamSize(fp,csv,"Shared RAM",sharedram);
  printRamSize(fp,csv,"Page Size",pagesize);


  printPeakCPU(fp,csv);
  printMinMaxSensors(fp,csv,num_temps,temp_names,min_temps,max_temps);

  if(errors == 0L)
    {
      sprintf(string,"PASSED, ran for %0.2f seconds\n\n",etime);
      printTranscript(string,TRUE);
    }
  else
    {
      sprintf(string,"FAILED, %ld errors\n\n",errors);
      printTranscript(string,TRUE);
    }
  printBankErrorSummary(fp,csv);

  if(fp)fprintf(fp,"*********************************** Summary across all thread counts ***********************************\n");
  if(csvt[0])fprintf(csv,"Summary across all thread counts\n");
  writeResults(fp,csvt[0],0,&final_results[0],ran,finfo);

  thread_base = 0;
  for(i=1;i<=threads;i++)
    {
      thread_limit = thread_base + i;
      if(thread_limit < MAX_THREADS)
	{
	  if(fp)fprintf(fp,"\n*********************************** Summary for thread count: %ld ***********************************\n",i);
	  writeResults(fp,NULL,i,&final_results[i],ran,finfo);
	  if(csvt[i])
	    {
	      fprintf(csvt[i],"\nSummary for thread count: %ld\n",i);
	      writeResults(NULL,csvt[i],i,&final_results[i],ran,finfo);
	    }
	}
      thread_base = thread_limit + 1;
    }
  closeTranscript();
  if(fp)fclose(fp);
  if(csv)fclose(csv);
  sync();
  closePageMap();
  munlock(mem,arraysize);
  free(mem);
  deleteDirs(MAX_THREADS,finfo);
  free(csvt);
}

